/*******************************************************************************

    METAL DETECTOR

    COM INDICAO PARA METAL FERROSO E NO FERROSO POR MEIO DE 2 LEDS

    C/ PIC16F676


    AUTOR: CLAUDIO LARIOS

    DATA:25/02/2024


    USO DIDTICO
    ESTE ARQUIVO  PARTE INTEGRANTE DO BLOG PICSOURCE

*******************************************************************************/
#include <16F676.h>
#device adc=10
#use delay(clock=20000000,restart_wdt)
#fuses WDT,HS, NOCPD, NOPROTECT, MCLR, PUT, BROWNOUT
//bytes
#byte  trisa = 0x85
#byte  porta  = 0x05
#byte  trisc = 0x87
#byte  portc  = 0x07
#byte  ansel  = 0x91
#byte  cmcon  = 0x19
#byte  vrcon  = 0x99
#byte  tmr0   = 0x01
#byte  tmr1h  = 0x0f
#byte  tmr1l  = 0x0e
#byte  t1con  = 0x10
#byte  pir1   = 0x0c
#byte  pie1   = 0x8c
#byte  intcon = 0x0b
#byte  option_reg = 0x81
#byte  pclath =  0x0a
#byte  pcl    =  0x02

#bit   t1if      = 0x0c.0
#bit   gie       = 0x0b.7
#bit   peie      = 0x0b.6
#bit   t0ie      = 0x0b.5
#bit   t0if      = 0x0b.2
#bit   cmif      = 0x0c.3 //flag de int. do comparador

#bit   lednf     = 0x07.1   //pino 9 verde
#bit   ledf      = 0x07.0   //pino 10 azul
#bit   led1      = 0x07.5   //pino 5
#bit   led2      = 0x07.4   //pino 6
#bit   led3      = 0x07.3   //pino 7
#bit   led4      = 0x07.2   //pino 8
#bit   buzzer    = 0x05.0 //pino 3 buzzer
//#bit sw_cal  = 0x05.3 //pino 4 boto de calibrao

//entrada de sinal da bobina osciladora detetora no pino 6 (cin-)
//sendo cin+ do comparador interno na vref interna (vrcon)
//oscilador externo ligado no pino 2 de 20 mhz com transistor e xtal
//pino 5 tem sada do comparador interno (ajuste freq entre 18 e 22 khz para operar)

//==============================================================================

//variveis globais
int1 flag_nl_ok=0;//flag indicador de nova leitura do timer 1 durante interrupo do tmr0
int1 flag_g1=0;//flag geral 1
int1 flag_g2=0;//flag geral 2
int1 flag_fnf=0;//flag indicador de metal ferroso ou no ferroso
int1 flag_ovft0=0;//flag auxiliar indicador de ocorrncia de estouro no timer 0
//unsigned int8 a,b;
unsigned int8 salva_tmr0=0,salva_tmr1h, salva_tmr1l,cnt1=0,cnt2=0,cnt_prox=0;
unsigned int16 new_tmr1,old_tmr1a=0,old_tmr1b=0,old_tmr1c=0,dif=0,old_dif=0,cnt_it=0;

/************************************************************************************
  Rotina de interrupo do timer 0
  Objetivo da rotina: Obter o valor da frequncia do oscilador na entrada (pino 6)
  do comparador de tenso. As alteraes entre as medies sero usadas para deter-
  minar a presena de metal ferroso  ou no ferroso.
/***********************************************************************************/

#INT_TIMER0

void timer0_isr(){
int8 a,b;
    t1con=0; //desliga timer 1
    a=tmr1h;
    b=tmr1l; 
    salva_tmr1h=a;//salva medio para uso nas rotinas
    salva_tmr1l=b;
    tmr1l=0;//zera timer1 h e timer 1l para nova medio
    tmr1h=0;//parte h=0  
    t1con=5; //religa timer 1
    tmr0=salva_tmr0;//recarga do timer 0
    flag_nl_ok=1;
}
/************************************************************************************
  Rotina de delay varivel em microssegundos.
  Objetivo da rotina: Obter delay de tempos reduzidos com restart do wdt.
/***********************************************************************************/

void var_delay_us(unsigned int8 v){

  if(v>=3)v=v-3; else return;
  while(v){
   v--;
   RESTART_WDT();
  };
}

/************************************************************************************
  Rotina de delay fixa em 10 milissegundos
/***********************************************************************************/

void delay10ms(){
  delay_ms(10);
}
/************************************************************************************
  Rotina de delay fixa em 100 milissegundos
/***********************************************************************************/
void delay100ms(){
   delay_ms(100);
}

/************************************************************************************
  Rotina de gerao de som
  Objetivo da rotina: Obter sons (apitos) com durao determinada por 'd' e frequncia
  sonoro determinada por 'e'.
/***********************************************************************************/

void gera_beep(unsigned int8 d,unsigned int8 e){
  unsigned int8 c;


  for(c=0;c<d;c++){
     buzzer=1;
     var_delay_us(e);
     buzzer=0;
     var_delay_us(e);
  }//for(c=0;c<=d;c++){
 }//void gera_beep(d,e){

/************************************************************************************
  Rotina de gerao de efeito sonoro.
  Objetivo da rotina: Gerar indicao de erros no  hardware por meio de beeps.
  Sendo o efeito sonoro comeando com beep longo, um delay 1 seg , mais outros
  beeps cuja quantidade indica qual  o erro, seguido de delay de 3 segs.

  1 beep = oscilador com frequncia abaixo de 18 khz
  2 beeps = frequncia acima de 22 khz
  4 beeps = oscilador morto (nota 1)


Nota1. com led verde piscando 12 vezes + 12 apitos = pino 6 sem sinal (osc. morto)
/***********************************************************************************/
void beep_erros(unsigned int8 n){
unsigned int8 d;

  for(d=0;d<8;d++){  // beep longo
   gera_beep(255,255);
  }
  for(d=0;d<10;d++){//1 segundo de delay
    delay100ms();
  }
  for(d=0;d<n;d++){
    gera_beep(230,150);
    delay100ms();
    delay100ms();
    delay100ms();
    delay100ms();
  }
  for(d=0;d<30;d++){//delay de 3 segundos
   delay100ms();
  }
  pclath=0;//pagina 0
  pcl=0;//reseta
}//beep_erros()

/************************************************************************************
  Rotina de indicao de presena de metais no-ferrosos
  Objetivo da rotina: gerar som distinto para metal no ferroso e acender led nf (verde)
/***********************************************************************************/

void nonferrous(){
unsigned int8 a;

  lednf=1; //liga led nf
  for(a=0;a<=160;a++){//lao para tempo de 40 mseg
    buzzer=1;
    var_delay_us(125);//freq ~ 4 khz ( nota B6)
    buzzer=0;
    var_delay_us(125);//130//110
  }
  delay10ms();
  //lednf=0;//desliga led nf
  portc=0;
}

/************************************************************************************
  Rotina de indicao de presena de metais ferrosos
  Objetivo da rotina: gerar som distinto para metal ferroso e acender led f (azul)
/***********************************************************************************/
void ferrous(){
unsigned int8 a;

  ledf=1;//liga led ferrous
  for(a=0;a<=80;a++){//lao para tempo de 40 mseg
    buzzer=1;
    var_delay_us(250);//freq ~ 2 khz (nota B5)
    buzzer=0;
    var_delay_us(250);
  }
  delay10ms();
 // ledf=0;//desliga led ferrous
  portc=0;
}
/************************************************************************************
  Rotina de ajuste dinmico do valor de comparao de leituras do timer 1
  Objetivo da rotina: Ajustar o valor de comparao compensando pequenas variaes
  da frequncia do oscilador. Isto minimiza indicaes erraticas constantes se no
  tivesse esta compensao. Necessita de 255 iteraes para a leitura de comparao
  ser alterada em um ponto. Em outras palavras o valor de comparao 'tende'
  lentamente a ser igual a leitura atual do timer 1.
/***********************************************************************************/

void dinamic_adjust_value(){

  RESTART_WDT();
  if(old_dif > dif) old_tmr1b=old_tmr1a;//se diferenca anterior  maior que diferena atual
  old_dif=dif;//diferena anterior fica com diferena atual

  if(old_tmr1c>new_tmr1)old_tmr1c--;
  if(old_tmr1c<new_tmr1)old_tmr1c++;//se valor anterior timer1 c for menor que nova leitura timer1

  cnt1++;
  if(cnt1>15){//rotina realizada a cada 15 iteraes
    cnt1=0;
    cnt2++;
    if(old_tmr1c>old_tmr1b)old_tmr1b++;//c>b=b++
    if(old_tmr1c<old_tmr1b)old_tmr1b--;//c<b=b--

  }
  if(cnt2>15){//rotina realizada a cada 255 iteraes
    cnt2=0;
    if(old_tmr1a>old_tmr1b)old_tmr1a--;//a>b=a--
    if(old_tmr1a<old_tmr1b)old_tmr1a++;//a<b=a++
  }
}//dinamic_adjust_value(){



/************************************************************************************
  Rotina de gerao de som
  Objetivo da rotina:
  Gerar beep de som caracteristico (ta-r-r)
/***********************************************************************************/
void efeito_sonoro_taranra(){
  unsigned int8 a;

  portc=0; //leds apagados
  for(a=0;a<3;a++){//gera beep 1
  gera_beep(100,240);//n. rep/periodo da nota
  }
  for(a=0;a<3;a++){
  delay10ms();//delay de 30 ms
  }
  for(a=0;a<3;a++){//gera beep 2
  gera_beep(200,120);//n. rep/periodo da nota
  }
  for(a=0;a<3;a++){
  delay10ms();//delay de 30 ms
  }
  for(a=0;a<3;a++){///gera beep 3
  gera_beep(120,200);//n. rep/periodo da nota
  }
}
/************************************************************************************
  Rotina de deteco de metal
  Objetivo da rotina:
  1) comparar as diferenas de leitura anterior e atual
  2) decidir se leitura  suficiente para acionar led/beep
  3) obter um delay proporcional a amplitude da diferena nas leituras:
     (obtendo o maior expoente da base 2 da diferena de leituras do timer 1
      div/4 e usando um contador de grau de proximidade de 3 nveis)
  4) decidir qual led acender (ferroso ou no-ferroso) e frequncia do beep
  5) acionar o bargraph de 4 leds conforme a proximidade.
/***********************************************************************************/

void detecta_metal(){
unsigned int16 aa,bb;
unsigned int8 c,proximity;
#define N_ITERACOES  12000

  RESTART_WDT();
  flag_nl_ok=0;//reseta flag
  while(!flag_nl_ok);//aguarda ficar '1' indicando nova leitura da frequencia osc.
  new_tmr1=(unsigned int16) salva_tmr1h<<8 | salva_tmr1l;


  if(new_tmr1>old_tmr1a){
    dif=new_tmr1 - old_tmr1a;//frequencia nova maior que anterior
    flag_fnf=0;//indica metal no ferroso
  }
 else if(new_tmr1==old_tmr1a){//sem alterao da frequncia (sem presena de metal no sensor)
   dif=0;
   flag_fnf=0;
 }
 else if(new_tmr1<old_tmr1a){//frequencia nova menor que anterior
    dif=old_tmr1a - new_tmr1;
    flag_fnf=1;//indica metal ferroso
 }


 if(dif<3){    //se diferena de leituras forem menor que 3
   cnt_prox=1;
   cnt_it++;  //incrementa contador de iteraes

   if(cnt_it>N_ITERACOES){  //testa se chegou a 12000 iteraes
    cnt_it=0;        //sim,zera contador de iteraes
    efeito_sonoro_taranra();    //chama rotina de som caracteristica
   }
   dinamic_adjust_value();  //rotina de autoajuste do valor de comparao da freq do osc.ext
 }//if(dif<3){


 else{//else - if(dif<3){
 //se diferena de leituras forem maior ou igual a 3
   cnt_it=0;

   //rotina de 3 nveis de indicao de aproximao
   // Objetivo: obter retardo proporcional ao valor da diferena de leitura
   //alterando a velocidade do beep e piscada dos led f/nf em 3 nveis
    aa=dif;
    bb=0x2000;
    for(c=0;c<14;c++){// em 'c' se obter um valor de 0 a 13
      RESTART_WDT();
      if(aa & bb)break;
      bb=bb>>1;
    }
    proximity=c>>2; //'c' ser dividido por 4 para obter apenas 3 nveis de visualizao


   if(proximity>=cnt_prox){//verifica se  a vez de acionar a sada conforme proximidade
     cnt_prox++;
   }// if(proximity>cnt_prox){
   else{
     cnt_prox=1;
     //obtem o efeito de bargraph com os leds 1 a 4
     switch (proximity){
      case 0: led4=1;
      case 1: led3=1;
      case 2: led2=1;
      case 3: led1=1;
   //   default: portc=0;//desliga todos leds
     }

     if(flag_fnf==1){
       nonferrous();//apita e acende led nonferrous (led verde)
     }
     else{
       ferrous();//apita e acende led ferrous (led azul)
     }

   }//else -  if(proximity>cnt_prox){

 }//else - if(dif<3){

}//void detecta_metal(){

/************************************************************************************
  Rotina de auto ajuste da tenso de referencia interna do pic.
  Funcionamento:
   A rotina verifica se houve mudana na contagem do timer 0 e caso isto no
   tenha ocorrido, incrementa o valor de VRCON at que isto ocorra.
   So 14 passos possveis. Depois de esgotado este valor da uma indicao de
   erro.
/***********************************************************************************/

void auto_vref_adjust(){
unsigned int8 b,f,j,new_tmr0;

   peie=0;
   gie=0;
   flag_g1=0;
   j=1;
   f=14;

   for(b=0;b<14;b++){
    j++;
    vrcon= j | 0xa0;
    nonferrous();
    delay100ms();
    delay100ms();
    tmr0=0;
    t0if=0;//reseta flag over flow do timer0
    delay10ms();
    flag_ovft0=0;
    if(t0if)flag_ovft0=1;//houve over flow do timer0
    new_tmr0=tmr0;
  //  flag_fnf=flag_ovft0;
    salva_tmr0=new_tmr0;//pega valor do timer 0 salvo


    if((new_tmr0==0)&& (flag_ovft0==0)){
     f--;
    }

    else{//e
     if(250<new_tmr0){beep_erros(2);}//freq>22khz, vai daqui para inicio do programa
     else{//f
       if(flag_ovft0==1){beep_erros(2);}//freq>22khz, vai daqui para inicio do programa
       else{
         if(2<new_tmr0){
          if(149>=new_tmr0){beep_erros(1);}//freq< 18khz, vai daqui para inicio do programa
         }
         flag_g1=1;
       }//x

     }//else{//f
    }//else{//e


    if(flag_g1==1)break;  //sai do lao
    if(f==0) break;;//sai do lao

   }// for(b=0;b<14;b++)  //loop

   if(f==0){
     beep_erros(4); //sem sinal pino 6, apito longo + pausa + 4x e goto inicio
   }
   else{
    if(j<14){
      j++;
      j=j |0b10100000;
      vrcon=j;
      salva_tmr0 = 0xff - new_tmr0;
    }
   }
}

/************************************************************************************
  Rotina de setup inicial de registradores e ports
/***********************************************************************************/

void initial_config(){

   ansel=0;//desliga conversor AD (675 apenas)
   cmcon=7;//desliga comparadores
   trisc=0;//zera tudo
   trisa=0b111010;
   portc=0;
   porta=0;
   delay100ms();//delay total de 300 ms
   delay100ms();
   delay100ms();
   gie=0;//desliga interrupes
   peie=0;
   tmr0=0;//zera timer 0

   SETUP_WDT(WDT_18MS);//ajusta modo de operao do WDT
   RESTART_WDT();//restarta WDT
   t1con=5; //liga timer 1
   cmcon=3; //liga comparador modo 011
   delay_us(10);//delay de 10 us
   cmif=0; //reseta flag comparador
   vrcon=0xa1;//LOW RANGE = VCC/24  3,3/24= 0,1375V GP2(PINO5) SADA DE FREQ

}

/************************************************************************************
  Rotina de estabilizao do oscilador ligado ao pino 6 (do osc. sensor de metal)
  Objetivo da rotina: Aguardar a estabilizao das leituras do timer 1
  (sem presena de metais na bobina do sensor do oscilador)
/***********************************************************************************/

void oscillator_stability(){
unsigned int8 a,f,g,i;

  flag_g2=0;//reseta flag
  f=0;
  g=0;
  old_tmr1a=0;
  tmr1l=0; //limpa timer1 l
  tmr1h=0; //limpa timer1 h
  tmr1l=0; //limpa timer1 l
  t0if=0;//reseta flag over flow do timer 0
  t1if=0;//reseta flag over flow do timer 1
  t0ie=1;//liga interrupo do timer 0 overflow
  peie=1;//liga interrupo dos perifricos
  gie=1;//liga geral das interrupes
  //testa sw calibrar /reseta
  while(!flag_nl_ok);//aguarda ficar '1'

  old_tmr1a=(unsigned int16) salva_tmr1h<<8 | salva_tmr1l;

  do{
    RESTART_WDT();
    flag_nl_ok=0;//reseta flag
    while(!flag_nl_ok);//aguarda ficar '1'
    new_tmr1=(unsigned int16) salva_tmr1h<<8 | salva_tmr1l;


    if(new_tmr1 > old_tmr1a) old_tmr1a++;
    if(old_tmr1a > new_tmr1) old_tmr1a--;


    g++;
    if(g==0)beep_erros(3);//se houver mais de 255 iteraes acusa erro

    if(old_tmr1a == new_tmr1){
      f++;
      if(f>3)flag_g2=1;
    }
    else{
     f=0;
     old_tmr1a=new_tmr1;
    }
}while(flag_g2==0);  // do{

//salva cpias nos registradores
old_tmr1b=old_tmr1a;
old_tmr1c=old_tmr1a;

portc=0; //leds apagados
i=255;
do{  //produz breve som (apito) do grave para agudo (tuuuuiiiiii)
   buzzer=1;
   var_delay_us(i);
   buzzer=0;
   var_delay_us(i);
   var_delay_us(i);
   i--;
}while(i>100);

for(a=0;a<5;a++)delay100ms();//delay 500 ms
}

/*******************************************************************************
          ROTINA MAIN
*******************************************************************************/

void main() {

  initial_config(); //configuraes iniciais de ports e registradores
  auto_vref_adjust();//ajuste automtico da tenso de referncia do comparador analgico interno
  oscillator_stability(); //aguarda estabilizar o oscilador do sensor de metal

/*******************************************************************************
           LOOP PRINCIPAL DE REPETIO
*******************************************************************************/
   for(;;){

    detecta_metal();//rotina de trabalho para detectar metais ferrosos e no ferrosos

   }//for(;;)
}//void main()
